/*
** Module   : SubMenu.java
** Abstract : SubMenu abstract client implementation
**
** Copyright (c) 2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -----------------------------Description-----------------------------
** 001 VIG 20150116 Created initial version.
** 002 HC  20150323 Removed business logic from config classes.
** 003 VIG 20150318 Added getFocusedChild() method and overridden focusWorker() method.
**                  Added event processing, repaint, hide/show body methods.
** 004 VIG 20150618 Moved several methods from ChUI implementation.
*/
package com.goldencode.p2j.ui.client;

import java.util.List;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.Iterables;

/**
 * <code>SubMenu</code> client abstract implementation class. 
 */
public abstract class SubMenu<O extends OutputManager<?>>
extends AbstractContainer<O>
implements WidgetWithConfig<SubMenuConfig>
{
   /** Menu attribute container. */
   protected SubMenuConfig config = null; 

   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public SubMenu(WidgetId id, SubMenuConfig cfg)
   {
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, SubMenuConfig cfg)
   {
      config = ConfigManager.getInstance().getActiveConfig(id);
      if (config == null)
      {
         config = new SubMenuConfig(id);
      }
      ConfigManager.getInstance().replaceConfig(config, cfg);
   }   

   /**
    * Provide access to the attributes stored in the widget.
    * 
    * @return  Reference to attribute container.
    */
   @SuppressWarnings("unchecked")
   @Override
   public SubMenuConfig config()
   {
      return config;
   }
   
   /**
    * The method is called after the configuration associated with 
    * the implementor changes. The parameter points to a reference
    * representing the original configuration state. This reference
    * can be used to detect what configuration fields changed and so
    * optimize any processing related to the configuration change.
    * 
    * @param    beforeUpdate
    */
   @Override
   public void afterConfigUpdate(SubMenuConfig beforeUpdate)
   {
      super.afterConfigUpdateBase(beforeUpdate);
   }

   /**
    * Hides sub-menu body.
    */
   public void hideBody(boolean recursive)
   {
      Rectangle bodyRect = 
               new Rectangle(bodyLocation(), dimension(), screen().coordinates().baseUnits());
      EventManager.postEvent(new PaintEvent(this, bodyRect));
      
      config().displayed = false;
      
      if (recursive)
      {
         if (parent() instanceof SubMenu)
         {
            ((SubMenu<?>) parent()).hideBody(true);
         }
         else if (parent() instanceof Menu)
         {
            Menu<?> menu = (Menu<?>)parent();
            if (menu.config().popupOnly)
            {
               // ((Menu<?>) parent()).hideBody();
               ((Menu<?>) parent()).hide();
            }
            else
            {
               menu.repaint(); // menubar
            }
         }
      }
      
      // TODO depends on triggers whether triggers defined or not. To be tested.
      // setFocus(null);
   }
   
   /**
    * Process key event.
    * 
    * @param   event
    *          Key event to process.
    */
   @Override
   public void processKeyEvent(KeyInput event)
   {
      if (event.isConsumed())
      {
         return;
      }

      int key = event.keyCode();
      int actionCode = event.actionCode();
      
      Menu<?> menu = Menu.getMenu(this);
      if (menu.config().popupOnly)
      {
         if (actionCode == Keyboard.KA_CURSOR_DOWN || actionCode == Keyboard.KA_CURSOR_RIGHT)
         {
            ThinClient.nextFocus(this);
            parent().repaint();
            return;
         }
         else if (actionCode == Keyboard.KA_CURSOR_UP || actionCode == Keyboard.KA_CURSOR_LEFT)
         {
            ThinClient.previousFocus(this);
            parent().repaint();
            return;
         }
      }
      
      // menu-item choosing events processing
      if (actionCode == Keyboard.SE_VALUE_CHANGED ||
          actionCode == Keyboard.SE_CHOOSE)
      {
         hideBody(true);
         return;
      }
      
      if (key == Key.VK_ENTER || 
          (Menu.isMenubarElement(this) && actionCode == Keyboard.KA_CURSOR_DOWN) ||
          (!Menu.isMenubarElement(this) && actionCode == Keyboard.KA_CURSOR_RIGHT))
      {
         // TODO what if body is already displayed
         // In this case there will be focused item in the body, that will process this event.
 
         showBody();
         KeyInput e = new KeyInput(this, EventType.KEY_PRESSED, Keyboard.SE_MENU_DROP, false, true);
         EventManager.postEvent(e);
         KeyInput e1 = new KeyInput(this, EventType.KEY_PRESSED, Keyboard.SE_MENU_DROP, false, true);
         EventManager.postEvent(e1);
         
         parent().repaint();
         event.consume();
         return;
      }
      
      if ((Menu.isMenubarElement(this) && actionCode == Keyboard.KA_CURSOR_RIGHT) || 
          (!Menu.isMenubarElement(this) && actionCode == Keyboard.KA_CURSOR_DOWN))
      {
         if (!Menu.isMenubarElement(this) && actionCode == Keyboard.KA_CURSOR_DOWN && config().displayed == true)
         {
            setInitialFocus();
         }
         else
         {
            ThinClient.nextFocus(this);
            /*if (config().displayed)
            {
               hideBody();
            }*/
         }
         
         parent().repaint();
         return;
      }
      
      if ((Menu.isMenubarElement(this) && actionCode == Keyboard.KA_CURSOR_LEFT) || 
          (!Menu.isMenubarElement(this) && actionCode == Keyboard.KA_CURSOR_UP))
      {
         if (!Menu.isMenubarElement(this) && actionCode == Keyboard.KA_CURSOR_DOWN && config().displayed == true)
         {
            setInitialFocus();
         }
         else
         {
            ThinClient.previousFocus(this);
            /*if (config().displayed)
            {
               hideBody();
            }*/
         }
         
         parent().repaint();
         return;
      }
      
      // hiding sub-menu from menu-item: only current if LEFT or all if RIGHT
      if ((!Menu.isMenubarElement(this) && actionCode == Keyboard.KA_CURSOR_RIGHT) || 
          (!Menu.isMenubarElement(this) && actionCode == Keyboard.KA_CURSOR_LEFT))
      {
         boolean recursive = actionCode == Keyboard.KA_CURSOR_RIGHT;
         if (parent() instanceof SubMenu)
         {
            ((SubMenu<?>) parent()).hideBody(recursive);

            if (actionCode == Keyboard.KA_CURSOR_RIGHT)
            {
               SubMenu<?> topSubmenu = (SubMenu<?>) parent();
               while (topSubmenu.parent() instanceof SubMenu<?>)
               {
                  topSubmenu = (SubMenu<?>) topSubmenu.parent();
               }

               if (Menu.isMenubarElement(topSubmenu))
               {
                  ThinClient.nextFocus(topSubmenu);
               }
            }
            else if (Menu.isMenubarElement((SubMenu<?>) parent()))
            {
               ThinClient.previousFocus(parent());
            }
         } 
         else if (((Menu<?>) parent()).config().popupOnly) // pop-up menu
         {
            ((Menu<?>) parent()).hide();
         }
         return;
      }
         
      // mnemonic
      if (event.keyChar() != KeyInput.CHAR_UNDEFINED)
      {
         parent().processKeyEvent(event);
      }
      else
      {
         //super.processKeyEvent(event);
      }
   }
   
   /**
    * Request widget repainting.
    */
   @Override
   public void repaint()
   {
      super.repaint();
      if (parent() instanceof SubMenu)
      {
         parent().repaint();
      }
   }
   
   /**
    * Hide/show widget.
    * <p>
    * If the menu is being shown and not yet attached to a window, attached it to the default
    * window.
    * 
    * @param   visible
    *          New visible state.
    */
   @Override
   public void setVisible(boolean visible)
   {
      // SUB-MENU is not displayed (dropped down) by default
      // setDisplayed(visible);
      
      for(Widget<O> w : widgets())
      {
         w.setVisible(true);
      }

      super.setVisible(visible);
   }

   /**
    * Check if widget needs in cursor to be on when having the input focus.  By default this
    * method always returns <code>true</code> so it needs to be overwritten if other return value
    * is required for particular widget implementation class.
    * 
    * @return  <code>true</code> as the default value.
    */
   @Override
   public boolean forceCursorOnFocus()
   {
      return false;
   }

   /**
    * Check if the current widget can manage non-global system actions on its own, and not via the
    * enclosing event loop.
    * 
    * @return   <code>true</code> as the widget has its own event management.
    */
   @Override
   public boolean manageSystemActions()
   {
      return true;
   }

   /**
    * Gets sub-menu title width (with '->' if sub-menu is in other sub-menu or in popup menu)
    * 
    * @return   width of sub-menu title
    */
   public abstract double titleWidth();
   
   /**
    * Get location of SUB-MENU body, relative to container.
    * 
    * @return  body location.
    */
   protected abstract Point bodyLocation();
   
   /**
    * Calculates body dimension
    * 
    * @return  Dimension
    */
   protected abstract Dimension bodyDimension();
   
   /**
    * Get first focused sub-menu child (menu-iter or other sub-menu)
    * 
    * @return   MenuItem or SubMenu
    */
   protected Widget<?> getFocusedChild()
   {
      for(Widget<?> w : widgets())
      {
         if (w.hasFocus())
         {
            return w;
         }
      }
      
      return null;
   }

   /**
    * Search for the next or previous focusable widget in the container and
    * set focus on it.
    *
    * @param   direction
    *          Widget search direction.
    */
   protected void focusWorker(boolean direction)
   {
      List<Widget<O>> widgets = getWidgetsList();
      Iterable<Widget<O>> iter = (direction) ?
                                 Iterables.directFrom(widgets, focus()) :
                                 Iterables.reverseFrom(widgets, focus());
                                     
       for (Widget<O> widget : iter)
       {
          if (checkWidget(widget, direction))
             return;
       }
    
       // cyclic transition to the first or last element
       Widget<O> nextWidget = direction ? widgets.get(0) : widgets.get(widgets.size() - 1);
       checkWidget(nextWidget, direction);
   }

   /**
    * Shows this sub-menu body.
    */
   protected void showBody() 
   {
      config().displayed = true;
      Window<?> wnd = Window.resolveWindow(this);
      Rectangle bodyRect = 
               new Rectangle(bodyLocation(), /* dimension() */bodyDimension(), wnd.screen().coordinates().baseUnits());
      EventManager.postEvent(new PaintEvent(this, bodyRect));
   }

   /**
    * Sets initial focus on first traversable child.
    */
   private void setInitialFocus()
   {
      Widget<?> firstChild = null;
      for (Widget<?> w : widgets())
      {
         if (w.focusTraversable())
         {
            firstChild = w;
            break;
         }
      }
      
      FocusManager.setInitialFocus(firstChild);
   }
}
