/*
** Module   : SubMenuGuiImpl.java
** Abstract : SubMenuGuiImpl client implementation
**
** Copyright (c) 2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -----------------------------Description-----------------------------
** 001 VIG 20150525 Created initial version.
*/
package com.goldencode.p2j.ui.client.gui;

import java.awt.event.MouseEvent;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.SubMenuChuiImpl;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.GuiFontResolver.FontCache;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * <code>SubMenu</code> client abstract implementation class. 
 */
public class SubMenuGuiImpl
extends SubMenu<GuiOutputManager>
{
   /** Screen driver */
   private GuiDriver gd;

   /** Font resolver */
   private GuiFontResolver gf;
   
   /** True when mouse is over the flat button.  */
   private boolean mouseEntered = false;

   /** Pressed button state flag. */
   private boolean pressed = false;
   
   /** shadow color. */
   private ColorRgb rule3DShadow;

   /** highlighted color */
   private ColorRgb highlightedColor;
   
   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public SubMenuGuiImpl(WidgetId id, SubMenuConfig cfg)
   {
      super(id, cfg);
   }

   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, SubMenuConfig cfg)
   {
      super.initialize(id, cfg);
      
      gd = (GuiDriver) OutputManager.getDriver();
      
      Widget<?> parent = ThinClient.getInstance().getWidget(cfg.frameId);
      Window<?> window = parent.window();
      
      gf = new GuiFontResolver(window, gd, config(), false);
      
      rule3DShadow       = gd.getSysColor(SysColor.COLOR_3DSHADOW);
      highlightedColor   = gd.getSysColor(SysColor.COLOR_MENUHILIGHT);
   }
   
   /**
    * Lay out container. Set children locations.
    */
   @Override
   public void doLayout()
   {
      super.doLayout();
      
      CoordinatesConversion c = this.ancestor().screen().coordinates();
      
      Widget<GuiOutputManager>[] w = widgets();
      int x = 0; // this.physicalLocation.x;
      int y = 0; // this.physicalLocation.y;
      
      for(int i=0; i<w.length; i++)
      {
         w[i].setPhysicalLocation(x, y + i * 25); // TODO 30 is a fixed menu item height
         w[i].setLocation(c.widthFromNative(x), c.heightFromNative(y + i * 25));

         if(w[i] instanceof SubMenu)
         {
            ((SubMenu<?>)w[i]).doLayout();
         }
      }
   }
   
   /**
    * Draw the peer container. The methods calls the draw() method for
    * ScrollPane instance for the peer container. In addition the required
    * clipping rectangle is calculated for ChildFrame containers because they
    * must be rendered inside the parent container.
    * <p>
    * Drawing for the redirected terminal differs. Labels/headers are drawn
    * only once per page, instead of the usual once per menu.
    */
   @Override
   public void draw()
   {
      final GuiColorResolver gce = new GuiColorResolver(gd, config(), 
               SysColor.COLOR_MENU, 
               SysColor.COLOR_BTNTEXT);
      gce.colors();
      
      final FontCache fc = gf.font();
      
      final NativePoint p = physicalLocation();
      final NativeDimension dim = physicalDimension();
      NativeRectangle clip = clipRectangle(p, dim);
      
      gd.draw(p, clip, new Runnable()
      {
         @Override
         public void run()
         {
            Window<?> wnd = drawTitleRectangle();
            CoordinatesConversion coords = wnd.screen().coordinates();

            drawTitle(fc, gce, wnd);
            
            if (config().displayed)
            {
               final NativePoint bodyLoc = coords.pointToNative(bodyLocation());
               final NativeDimension bodyDim = coords.dimensionToNative(bodyDimension());
               
               // panel
               NativeRectangle clip = clipRectangle(bodyLoc, bodyDim);
               gd.draw(bodyLoc, clip, new Runnable()
               {
                  @Override
                  public void run()
                  {
                     gd.setColor(gce.bgColor);
                     gd.fillRect(0, 0, bodyDim.width, bodyDim.height);
                     gd.draw3DRect(0, 0, bodyDim.width, bodyDim.height, true);
                     
                     SubMenuGuiImpl.super.draw();
                  }
               });
            }
         }

         /**
          * Draws rectangle around submenu title
          * 
          * @return  sub-menu window.
          */
         private Window<?> drawTitleRectangle()
         {
            Window<?> wnd = Window.resolveWindow(SubMenuGuiImpl.this);
            CoordinatesConversion coord = wnd.screen().coordinates();

            int titleHeight = coord.heightToNative(SubMenuGuiImpl.this.titleHeight()) + 5 + 5;
            int titleWidth = coord.widthToNative(SubMenuGuiImpl.this.titleWidth()) + 5 + 5;
            
            if (Menu.isMenubarElement(SubMenuGuiImpl.this))
            {
               if (SubMenuGuiImpl.this.pressed)
               {
                  gd.setColor(rule3DShadow);
                  gd.draw3DRect(2, 2, titleWidth - 2, titleHeight - 4, false);
               } 
               else if (SubMenuGuiImpl.this.mouseEntered)
               {
                  gd.setColor(rule3DShadow);
                  gd.draw3DRect(2, 2, titleWidth - 2, titleHeight - 4, true);
               } 
               else 
               {
                  gd.setColor(gce.bgColor);
                  gd.drawRect(2, 2, titleWidth - 2, titleHeight - 4);
               }
            }
            else
            {
               if (SubMenuGuiImpl.this.mouseEntered)
               {
                  int width = coord.widthToNative(parent().width());
                  gd.setColor(highlightedColor);
                  gd.fillRect(2, 2, width - 2, dim.height - 2);
               } 
            }

            return wnd;
         }
      });
   }
   
   /**
    * Get widget width.
    * 
    * @return  widget width.  
    */
   @Override
   public double width()
   {
      double w = titleWidth();
      
      if (config().displayed)
      {
         if (Menu.isMenubarElement(this))
         {
            w = Math.max(w, getMaxSubMenuItemWidth(this, true));
         }
         else
         {
            w += getMaxSubMenuItemWidth(this, true);
         }
      }
      
      // borders
      return w + Window.resolveWindow(this).screen().coordinates().widthFromNative(5 + 5);
   }

   /**
    * Get sub-menu height.
    * 
    * @return  sub-menu height.
    */
   @Override
   public double height()
   {
      Window<?> wnd = Window.resolveWindow(this);
      int h = FontManager.getFontHeight(wnd, gf.resolveFontNum(), gd);
      double dh = wnd.screen().coordinates().heightFromNative(h + 5 + 5);
      
      // TODO sum of items heights
      
      if (config().displayed)
      {
         dh += SubMenuGuiImpl.getMaxSubMenuHeight(this, true);
      }
      
      return dh;
   }

   /**
    * Hides sub-menu body.
    */
   @Override
   public void hideBody(boolean recursive)
   {
      this.pressed = false;
      super.hideBody(recursive);
   }
   
   /**
    * Gets sub-menu title width (with '->' if sub-menu is in other sub-menu or in popup menu)
    * 
    * @return   width of sub-menu title
    */
   @Override
   public double titleWidth()
   {
      Window<?> wnd = Window.resolveWindow(this);
      CoordinatesConversion coords = wnd.screen().coordinates();
      
      int w = FontManager.getTextWidth(config().title, wnd, gf.resolveFontNum(), gd);
      
      if (!Menu.isMenubarElement(this))
      {
         // an arrow
         w += 4 + 5;
      }
      
      return coords.widthFromNative(w);
   }
   
   /**
    * Gets sub-menu title height
    * 
    * @return   height of sub-menu title
    */
   public double titleHeight()
   {
      Window<?> wnd = Window.resolveWindow(this);
      int h = FontManager.getFontHeight(wnd, gf.resolveFontNum(), gd);
      return wnd.screen().coordinates().heightFromNative(h);
   }
   
   /**
    * Notification of a mouse clicked event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      // and simulate ENTER key to select it
      KeyInput ki = new KeyInput(this, EventType.KEY_PRESSED, Key.VK_ENTER);
      EventManager.postEvent(ki);
   }
   
   /**
    * Notification of a mouse entered event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseEntered(MouseEvent e)
   {
      boolean oldVal = mouseEntered; 
      mouseEntered = true;
      
      if (!oldVal)
      {
         requestFocus();
         parent().repaint();
      }
   }

   /**
    * Notification of a mouse exited event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseExited(MouseEvent e)
   {
      boolean oldVal = mouseEntered; 
      mouseEntered = false;
      
      if (oldVal)
      {
         parent().repaint();
      }
   }

   /**
    * Notification of a mouse pressed event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      boolean oldVal = pressed; 
      pressed = true;
      
      if (!oldVal)
      {
         parent().repaint();
      }
   }

   /**
    * Notification of a mouse released event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   /*
   @Override
   public void mouseReleased(MouseEvent e)
   {
      boolean oldVal = pressed; 
      pressed = false;
      
      if (oldVal)
      {
         parent().repaint();
      }
   }
   */

   /**
    * Notification of a mouse dragged event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseDragged(MouseEvent e)
   {
      // get coords of the drag event related to button
      NativePoint mp = translate(this, e);
      // get button location and size
      NativePoint p = physicalLocation();
      NativeDimension d = physicalDimension();
      // this can be rare case when we leave the button with holded mouse key, the button must
      // be released, the mouseExited event will not be delivered to button until drag button
      // released even if the pointer is far away from button accepted press event
      if ((mp.x < p.x || mp.x > (p.x + d.width) || mp.y < p.y || mp.y > (p.y + d.height)) &&
          pressed)
      {
         pressed = false;
         mouseEntered = false;
         parent().repaint();
      }
   }
   
   /**
    * Find the widget positioned just bellow the specified mouse position (in physical units).
    * Each container should search for own widgets. Because widgets are relative positioned 
    * inside containers the container physical position is adjusted before search.  
    * 
    * @param    p
    *           The mouse physical location.
    * 
    * @return   A widget over which the mouse is positioned or <code>null</code> if none found.
    */
   @Override
   public Widget<GuiOutputManager> findMouseSource(NativePoint p)
   {
      if (!config().displayed)
      {
         return null;
      }
      
      CoordinatesConversion coord = Window.resolveWindow(this).screen().coordinates();
      NativePoint tp = new NativePoint(p.x, p.y - coord.heightToNative(this.titleHeight()));
      
      return super.findMouseSource(tp);
   }
   
   /**
    * Get children location for one child.
    * 
    * @param   child
    *          Child component.
    *          
    * @return  Point - parents location
    */
   @Override
   public Point childrenLocation(Widget<GuiOutputManager> child)
   {
      /*
      Window<?> wnd = Window.resolveWindow(this);
      Point bodyLoc = bodyLocation();
      NativePoint p = wnd.screen().coordinates().pointToNative(bodyLoc);
      int fontHeight = FontManager.getFontHeight(wnd, gf.resolveFontNum(), gd);
      p.y += fontHeight + 5 + 5;  // under the title
      return wnd.screen().coordinates().pointFromNative(p);
      */
      return bodyLocation(); // TODO add borders?
   }
   
   /**
    * Calculates a max height of specified Menu or SubMenu.
    * Analyzes nested dropped-down sub-menus.
    * 
    * @param   menu
    *          Menu or SubMenu.
    *          
    * @param   includeNested
    *          <code>true</code> process nested sub-menus.
    * 
    * @return  a height of menu.
    */
   public static double getMaxSubMenuHeight(AbstractContainer<?> menu, boolean includeNested)
   {
      Window<?> wnd = Window.resolveWindow(menu);
      GuiDriver gd = (GuiDriver) OutputManager.getDriver();
      GuiFontResolver gf = new GuiFontResolver(wnd, gd, (BaseConfig)menu.config(), false);
      int itemHeight = FontManager.getFontHeight(wnd, gf.resolveFontNum(), gd) + 5 + 5;
      double itemHeightDbl = wnd.screen().coordinates().heightFromNative(itemHeight);
      
      Widget<?>[] widgets = menu.widgets();
      double height = widgets.length * itemHeightDbl;
      
      if (includeNested)
      {
         for (int i = 0; i < widgets.length; i++)
         {
            if (widgets[i] instanceof SubMenu)
            {
               SubMenu<?> sm = (SubMenu<?>) widgets[i];
               if (sm.config().displayed)
               {
                  height = Math.max(height, sm.height() + (i + 1) * itemHeightDbl); // this.y + title height
               }
            }
         }
      }
      
      return height;
   }
   
   /**
    * Calculates max menu/sub-menu item width
    * 
    * @param    menu
    *           MENU or SUB-MENU
    * @param    includeNested
    *           <code>true</code> means, that we should process nested sub-menus, 
    *           <code>false</code> otherwise.          
    *           
    * @return   max width.
    */
   public static double getMaxSubMenuItemWidth(Container<?> menu, boolean includeNested)
   {
      double width = 0; 
      for(Widget<?> w : menu.widgets())
      {
         AbstractWidget<?> aw = (AbstractWidget<?>) w;
         
         // ignore SKIP or RULE
         if (aw instanceof MenuItem && !((MenuItem<?>)aw).isNormal())
         {
            continue;
         }
         
         double itemWidth;
         if (aw instanceof SubMenu)
         {
            itemWidth = ((SubMenu<?>)aw).titleWidth();
            if (includeNested)
            {
               itemWidth += aw.width();
            }
         }
         else
         {
            itemWidth = aw.width();
         }
         
         if (width < itemWidth)
         {
            width = itemWidth; 
         }
      }
      return width;
   }
   
   /**
    * Get location of SUB-MENU body, relative to container.
    * 
    * @return  body location.
    */
   @Override
   protected Point bodyLocation()
   {
      Window<?> wnd = Window.resolveWindow(this);
      CoordinatesConversion coord = wnd.screen().coordinates();
      NativePoint p = new NativePoint(0, 0);
      if (!Menu.isMenubarElement(this))
      {
         p.x += coord.widthToNative(parent().width());
      }
      else
      {
         int fontHeight = FontManager.getFontHeight(wnd, gf.resolveFontNum(), gd);
         p.y += fontHeight + 5 + 5;  // under the title
      }
      return coord.pointFromNative(p); 
   }
   
   /**
    * Calculates body dimension
    * 
    * @return  Dimension
    */
   @Override
   protected Dimension bodyDimension()
   {
      double height = SubMenuGuiImpl.getMaxSubMenuHeight(this, false);
      double width = SubMenuGuiImpl.getMaxSubMenuItemWidth(this, false);
      CoordinatesConversion coords = Window.resolveWindow(this).screen().coordinates();
      NativeDimension dim = coords.dimensionToNative(new Dimension(width, height));
      dim.width += 7 + 5; // borders
      dim.height += 5 + 5;
      return Window.resolveWindow(this).screen().coordinates().dimensionFromNative(dim);
   }

   /**
    * Draws menu-item title, drawing mnemonic character as underlined.
    * 
    * @param   fc
    *          Font cache
    * @param   gce
    *          Color resolver
    * @param   wnd
    *          Window owning this sub-menu
    */
   private void drawTitle(final FontCache fc, final GuiColorResolver gce, final Window<?> wnd)
   {
      int fh = FontManager.getFontHeight(wnd, gf.resolveFontNum(), gd);

      gd.setColor(gce.fgColor);
      gd.setGuiFont(fc.font);
      gd.setFontStyle(fc.font.style);

      NativePoint tp = new NativePoint(7, 2 + fh);

      String title = config().title;

      // if & is not specified use the first letter as mnemonic
      int ampIdx = title.indexOf('&');
      boolean amp = ampIdx != -1; 
      if (!amp) ampIdx = 0;

      String leftPart = title.substring(0, ampIdx);
      gd.drawString(leftPart, tp.x, tp.y);

      tp.x += FontManager.getTextWidth(leftPart, wnd, gf.resolveFontNum(), gd);

      // "eat" the ampersand
      if (amp) ampIdx++;

      String mnem = title.substring(ampIdx, ampIdx + 1);
      gd.setFontStyle(FontStyle.UNDERLINE);
      gd.drawString(mnem, tp.x, tp.y);

      tp.x += FontManager.getTextWidth(mnem, wnd, gf.resolveFontNum(), gd);

      gd.setFontStyle(fc.font.style);
      String postFix = title.substring(ampIdx + 1);
      gd.drawString(postFix, tp.x, tp.y);
      
      tp.x += FontManager.getTextWidth(postFix, wnd, gf.resolveFontNum(), gd);
      
      if (!Menu.isMenubarElement(this))
      {
         tp.x += 5;
         gd.drawLine(tp.x, tp.y, tp.x, tp.y - 8);
         gd.drawLine(tp.x, tp.y, tp.x + 4, tp.y - 4);
         gd.drawLine(tp.x, tp.y - 8, tp.x + 4, tp.y - 4);
      }
   }
   
}
